/*
    c99tree - main
    Copyright (C) 2010-2015  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "c99tree-cc.h"
#include "tree_print.h"
#include "tree_load.h"
#include "revision.h"
#include "scope.h"
#include "gtx.h"

static int preprocessed = 0; /* detect direct calls and warn the user; treecc always issues -P */
extern int c99c_debug; /* bison does this */

/* --- error handling --- */
void c99c_werror(c99tree_cc_t *ctx, const char *s)
{
	fprintf(stderr, "c99c_werror: ");
	c99c_error(ctx, s);
}

void c99c_error(c99tree_cc_t *ctx, const char *s)
{
	c99c_error_fprintf(ctx, s, stderr);
	exit(2);
}

/* --- command line argument handling --- */
static void help(c99tree_cc_t *ctx, const char *prg, const char *topic)
{
	if (topic == NULL)
		topic = "";

	if (strcasecmp(topic, "format") == 0) {
		fprintf(stderr, "Available input formats:\n");
		fprintf(stderr, "  -tc         C source file (default)\n");
		fprintf(stderr, "  -tind       space indented tree\n");
#ifdef LOAD_XML
		fprintf(stderr, "  -txml       xml\n");
#endif
		fprintf(stderr, "\n");
		fprintf(stderr, "Available output formats:\n");
		fprintf(stderr, "  -Tind       space indented tree (default)\n");
		fprintf(stderr, "  -Tweb3      web3.0\n");
		fprintf(stderr, "  -Txml       xml (long style means separate closing tag)\n");
		fprintf(stderr, "  -Tplist     property list\n");
		fprintf(stderr, "  -Tjson      json\n");
		fprintf(stderr, "  -Tperl      perl\n");
		fprintf(stderr, "  -Tc         C source file\n");
		fprintf(stderr, "  -Tdot       digraph for dot (graphviz)\n");
		fprintf(stderr, "  -Tawk       awk script (function that builds the tree in an array)\n");

	}
	else if (strcasecmp(topic, "verbose") == 0) {
		fprintf(stderr, "Available verbose options (off by default, turned on by -v*):\n");
		fprintf(stderr, "  -vtypedef   expand typedef'd types when referenced\n");
		fprintf(stderr, "  -vall       turn on all of the above -v options\n");
	}
	else if (strcasecmp(topic, "location") == 0) {
		fprintf(stderr, "Location format is similar to the format string of printf(3),\n");
		fprintf(stderr, "with the following directives:\n");
		fprintf(stderr, "  %%f          file name (as referenced)\n");
		fprintf(stderr, "  %%p          path name (file name with path if looked up)\n");
		fprintf(stderr, "  %%l          start line of the token\n");
		fprintf(stderr, "  %%L          end line of the token\n");
		fprintf(stderr, "  %%c          start column of the token\n");
		fprintf(stderr, "  %%C          end column of the token\n");
		fprintf(stderr, "  %%e          macro expansion nesting level\n");
		fprintf(stderr, "  %%E          1 if line is local (not included)\n");
		fprintf(stderr, "  %%o          start byte offset of the token\n");
		fprintf(stderr, "  %%O          end byte offset of the token\n");
		fprintf(stderr, "  %%m          cpp #line macro location: file name\n");
		fprintf(stderr, "  %%M          cpp #line macro location: line\n");
		fprintf(stderr, "  %%y          cgram.y line number (requires compile time option)\n");
		fprintf(stderr, "  %%%%          a  plain %% character\n");

		fprintf(stderr, "\n");
		fprintf(stderr, "Any other character is copied as is.\n");
		fprintf(stderr, "For example, the default format is -L%s\n", ctx->loc_format);
	}
	else {
		fprintf(stderr, "c99tree - parse preprocessed C source file into a computer readable tree\n");
		fprintf(stderr, "Usage:\n");
		fprintf(stderr, " %s [-tformat] [-Tformat] [-ytypename] [options] [filename]\n", prg);
		fprintf(stderr, " %s --help [topic]\n", prg);
		fprintf(stderr, " %s -V\n\n", prg);
		fprintf(stderr, "  -tformat         input format (try --help format)\n");
		fprintf(stderr, "  -Tformat         output format (try --help format)\n");
		fprintf(stderr, "  -ytypename       register a new base type (multiple of this is allowed)\n");
		fprintf(stderr, "  -Ytypename       same as y, but make the new type replacable/redefinable\n");
		fprintf(stderr, "  -i               print nodes coming from include files\n");
		fprintf(stderr, "  -P               input has been preprocessed with c99tree-cpp\n");
		fprintf(stderr, "  -l               set output style 'long'\n");
		fprintf(stderr, "  -Lformat         change location format (see --help location)\n");
		fprintf(stderr, "  -d               debug bison parser\n");
		fprintf(stderr, "  -g gentrex_expr  run gentrex expression on the grammar tree after processing\n");
		fprintf(stderr, "  -v               set verbosity options (try --help verbose)\n");
		fprintf(stderr, "  filename         read input from filename instead of stdin\n");
		fprintf(stderr, "  --help           this help screen or help screen about topic\n");
		fprintf(stderr, "  -V               print svn revision and exit\n");

		if (!preprocessed) {
			fprintf(stderr, "\n");
			fprintf(stderr, "Note: c99tree-cpp shall be used for the preprocessing (for custom directives)\n");
		}
		fprintf(stderr, "\n");
	}

/*	fprintf(stderr, "\n(options marked with * are not fully functional)\n"); */
	exit(0);
}

static void output_format(c99tree_cc_t *ctx, const char *fmt)
{
	if (c99c_set_output_format(ctx, fmt) == NULL) {
		fprintf(stderr, "c99tree error: unkown output format %s\n", fmt);
		exit(3);
	}
}

static void set_verbose(c99tree_cc_t *ctx, const char *name)
{
	if (c99c_set_verbose(ctx, name) != 0) {
		fprintf(stderr, "Error: unknown verbose option name '%s'\n", name);
		exit(1);
	}
}

int main(int argc, char *argv[])
{
	char *cmd;
	int n;
	c99c_node_t *root;
	c99c_node_t node_base_type;
	char *input_format = "c";
	int local_only = 1;
	c99tree_cc_t ctx;
	char *filename = NULL;

	c99tree_cc_init(&ctx);

	memset(&node_base_type, 0, sizeof(c99c_node_t));
	node_base_type.type=T_TYPENAME;

	for(n = 1; n < argc; n++) {
		cmd = argv[n];
		if (*cmd == '-') {
			while(*cmd == '-') cmd++;
			switch(*cmd) {
				case 'h': help(&ctx, argv[0], argv[n+1]); break;
				case 'T': output_format(&ctx, cmd+1); break;
				case 't': input_format = cmd+1; break;
				case 'd': c99c_debug = 1; break;
				case 'l': ctx.format_long = 1; break;
				case 'L': ctx.loc_format = cmd+1; break;
				case 'v': set_verbose(&ctx, cmd+1); break;
				case 'V': printf("c99tree %s\n", svn_revision); return 0;
				case 'i': local_only = 0; break;
				case 'P': preprocessed = 1; break;
				case 'g': if (cmd[1] == 't') { n++; c99c_gtx_list_append(&ctx, argv[n]); } break;
				case 'y':
				case 'Y':
					{
						c99c_node_t *n = C99Nclone(&ctx, &node_base_type, INV_LOC_NONE);
						n->name = c99c_strclone(cmd+1);
						n->redefinable = (*cmd == 'Y');
						c99c_scope_add_typedef(&ctx, cmd+1, n);
						break;
					}
			}
		}
		else {
			if (filename != NULL) {
				fprintf(stderr, "c99tree-cc error: multiple filenames on command line (%s, %s).\n", filename, cmd);
				return 1;
			}
			filename = cmd;
		}
	}

	/* parse the file */
	root = c99c_tree_load(&ctx, filename, input_format);
	c99c_tree_postproc(root);

	c99c_gtx_run(&ctx, root);

	/* print the result */
	if (root != NULL) {
		void *nprint_ctx = NULL;
		if (ctx.Nprint_cb(&ctx, &nprint_ctx, "", root, PS_HEAD) == 0) {
			C99Nprint(&ctx, &nprint_ctx, " ", root, local_only);
			ctx.Nprint_cb(&ctx, &nprint_ctx, "", root, PS_TAIL);
		}
	}

	c99c_loc_file_close(&ctx);
	if (ctx.loc_current != NULL)
		fprintf(stderr, "Internal error: location stack got confused. Handling of #includes is broken.\n");

	/* free up memory */
	c99c_tree_load_cleanup(&ctx, input_format);

	C99Nfree(&ctx, root);
	c99tree_cc_uninit(&ctx);
	return 0;
}
