/*
    c99tree - execute gentrex scripts on the tree
    Copyright (C) 2015  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "gtx.h"

void *c99c_tree_child(void *user_ctx_, void *node_)
{
	c99c_node_t *node = node_;
	return node->children;
}

void *c99c_tree_parent(void *user_ctx_, void *node_)
{
	c99c_node_t *node = node_;

	return node->parent;
}

void *c99c_tree_next(void *user_ctx_, void *node_)
{
	c99c_node_t *node = node_;
	return node->next;
}

typedef enum node_field_e {
	NF_INVALID,
	NF_UID,
	NF_NAME,
	NF_TYPE,
	NF_LOC_FILE,
	NF_LOC_SLINE,
	NF_LOC_ELINE,
	NF_LOC_SCOL,
	NF_LOC_ECOL,
	NF_LOC_SOFFS,
	NF_LOC_EOFFS,
	NF_LOC_IS_LOCAL,
	NF_LOC_MACRO_LEVEL,
} node_field_t;

typedef struct op_cookie_s {
	node_field_t fld;
	int ival;
	intop_t iop;
} op_cookie_t;

static void *op_cookie_alloc(void *user_tree_ctx, gentrex_op_t opt, const char *tagname, const char *str_or_pattern, const char *custom_operator)
{
	op_cookie_t *ck;
	int ival;
	node_field_t fld = NF_INVALID;
	c99tree_cc_t *ctx = user_tree_ctx;

	ival = atoi(str_or_pattern);

	if (strcmp(tagname, "uid") == 0)
		fld = NF_UID;
	else if (strcmp(tagname, "name") == 0)
		fld = NF_NAME;
	else if (strcmp(tagname, "type") == 0) {
		fld = NF_TYPE;
		ival = C99Ntype(ctx, str_or_pattern);
	}
	else if (strncmp(tagname, "loc_", 4) == 0) {
		tagname += 4;
		if (strcmp(tagname, "file") == 0)
			fld = NF_LOC_FILE;
		else if (strcmp(tagname, "sline") == 0)
			fld = NF_LOC_SLINE;
		else if (strcmp(tagname, "eline") == 0)
			fld = NF_LOC_ELINE;
		else if (strcmp(tagname, "scol") == 0)
			fld = NF_LOC_SCOL;
		else if (strcmp(tagname, "ecol") == 0)
			fld = NF_LOC_SCOL;
		else if (strcmp(tagname, "soffs") == 0)
			fld = NF_LOC_SOFFS;
		else if (strcmp(tagname, "eoffs") == 0)
			fld = NF_LOC_EOFFS;
		else if (strcmp(tagname, "is_local") == 0)
			fld = NF_LOC_IS_LOCAL;
		else if (strcmp(tagname, "macro_level") == 0)
			fld = NF_LOC_MACRO_LEVEL;
	}
	
	if (fld == NF_INVALID) {
		fprintf(stderr, "Error: invalid gentrex tag (node field): '%s'\n", tagname);
		return NULL;
	}

	ck = malloc(sizeof(op_cookie_t));
	ck->fld = fld;
	ck->ival = ival;
	if (opt == GTX_OP_C_OP)
		ck->iop = intop_parse(custom_operator);

	return ck;
}

static void op_cookie_free(void *user_tree_ctx, gentrex_op_t opt, const char *tagname, const char *str_or_pattern, void *op_cookie, const char *custom_operator)
{
	free(op_cookie);
}

static int strmatch(const char *s1, const char *s2, int wild)
{
	if ((s1 == NULL) && (s2 == NULL))
		return 1;
	if ((s1 == NULL) || (s2 == NULL))
		return 0;

	if (wild)
		return strstr(s1, s2) != NULL;
	return !strcmp(s1, s2);
}

static int tag_match_(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *str, int wild)
{
	c99c_node_t *node = node_;
	op_cookie_t *ck = op_cookie;

	if (ck == NULL)
		return 0;

	switch(ck->fld) {
		case NF_UID:              return node->uid == ck->ival;
		case NF_NAME:             return strmatch(node->name, str, wild);
		case NF_LOC_FILE:         return strmatch(node->loc.file->fn, str, wild);
		case NF_LOC_SLINE:        return node->loc.sline == ck->ival;
		case NF_LOC_ELINE:        return node->loc.eline == ck->ival;
		case NF_LOC_SCOL:         return node->loc.scol == ck->ival;
		case NF_LOC_ECOL:         return node->loc.ecol == ck->ival;
		case NF_LOC_SOFFS:        return node->loc.soffs == ck->ival;
		case NF_LOC_EOFFS:        return node->loc.eoffs == ck->ival;
		case NF_LOC_IS_LOCAL:     return node->loc.is_local == ck->ival;
		case NF_LOC_MACRO_LEVEL:  return node->loc.macro_level == ck->ival;
		case NF_TYPE:             return node->type == ck->ival;
		case NF_INVALID:
			abort();
	}
	return 0;
}

static int tag_match(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *str)
{
	return tag_match_(user_tree_ctx, node_, op_cookie, tagname, str, 1);
}

static int tag_eq(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *str)
{
	return tag_match_(user_tree_ctx, node_, op_cookie, tagname, str, 0);
}


static int tag_c_op(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *str, const char *c_op)
{
	c99c_node_t *node = node_;
	op_cookie_t *ck = op_cookie;
	int ival;

	if (ck == NULL)
		return 0;

	switch(ck->fld) {
		case NF_NAME:
		case NF_LOC_FILE:
			ival = -1;
			break;
		case NF_UID:             ival = node->uid; break;
		case NF_LOC_SLINE:       ival = node->loc.sline; break;
		case NF_LOC_ELINE:       ival = node->loc.eline; break;
		case NF_LOC_SCOL:        ival = node->loc.scol; break;
		case NF_LOC_ECOL:        ival = node->loc.ecol; break;
		case NF_LOC_SOFFS:       ival = node->loc.soffs; break;
		case NF_LOC_EOFFS:       ival = node->loc.eoffs; break;
		case NF_LOC_IS_LOCAL:    ival = node->loc.is_local; break;
		case NF_LOC_MACRO_LEVEL: ival = node->loc.macro_level; break;
		case NF_TYPE:            ival = node->type; break;
		case NF_INVALID:
			abort();
	}

	intop_cmp(ival, ck->iop, ck->ival, return 0);

	return 0;
}

void c99c_gtx_init(c99tree_cc_t *ctx)
{
	memset(&ctx->api, 0, sizeof(ctx->api));
	ctx->api.user_map_ctx = &ctx->map;
	ctx->api.map_get = gentrex_map_az_get;
	ctx->api.map_set = gentrex_map_az_set;

	ctx->api.next   = c99c_tree_next;
	ctx->api.child  = c99c_tree_child;
	ctx->api.parent = c99c_tree_parent;

	ctx->api.op_cookie_alloc = op_cookie_alloc;
	ctx->api.op_cookie_free  = op_cookie_free;
	ctx->api.tag_eq    = tag_eq;
	ctx->api.tag_match = tag_match;
	ctx->api.tag_c_op  = tag_c_op;

	ctx->gtx_head = NULL;
	ctx->gtx_tail = NULL;
}

void c99c_gtx_list_append(c99tree_cc_t *ctx, char *arg)
{
	c99c_gtx_list_t *l;
	const char *err;
	char *s, *script, *name;

	for(s = arg; *s != '\0' && (isalnum(*s) || (*s == '_')) ;s++) ;

	if (*s == '=') {
		name = arg;
		script = s+1;
		*s = '\0';
		if (*name == '\0')
			name = NULL;
	}
	else {
		name = NULL;
		script = arg;
	}

	l = malloc(sizeof(c99c_gtx_list_t));
	l->name = name;
	l->script = gentrex_alloc(&ctx->api, script, &err);
	if (l->script == NULL) {
		fprintf(stderr, "Error compiling gentrex script '%s': %s\n", script, err);
		exit(1);
	}
	l->next = NULL;

	if (ctx->gtx_tail != NULL) {
		ctx->gtx_tail->next = l;
		ctx->gtx_tail = l;
	}
	else
		ctx->gtx_head = ctx->gtx_tail = l;
}

void c99c_gtx_run(c99tree_cc_t *ctx, c99c_node_t *root)
{
	c99c_gtx_list_t *l;
	c99c_node_t *res_root;
	int n;

	if (ctx->gtx_head == NULL)
		return;

	ctx->api.user_tree_ctx = ctx;

	/* save the initial map for restoring in each iteration */
	memcpy(&ctx->map_init, &ctx->map, sizeof(ctx->map));

	res_root = C99Nalloc(ctx, T_GENTREX, NULL, NULL);
	root->next->next = res_root;

	for(l = ctx->gtx_head, n = 0; l != NULL; l = l->next, n++) {
		c99c_node_t *scr;
		const char *name;
		char name_buff[64];
		
		name = l->name;
		if (name == NULL) {
			name = name_buff;
			sprintf(name_buff, "gtx_%d", n);
		}
		scr = C99Nalloc(ctx, T_GENTREX_RESULT, c99c_strclone(name), NULL);
		for(;;) {
			c99c_node_t *res, *lnk;
			char uids[32];
			int m;

			/* restore the initial map for each subsequent query */
			memcpy(&ctx->map, &ctx->map_init, sizeof(ctx->map));

			res = gentrex_execute(l->script, root);
			if (res == NULL)
				break;

			sprintf(uids, "%d", res->uid);
			lnk = C99Nalloc(ctx, T_SYMLINK, c99c_strclone(uids), NULL);
			C99Nchappend(ctx, scr, lnk);
			lnk->parent = scr;

			for(m = 0; m < 26; m++) {
				if (ctx->map.map[m].straight != NULL) {
					char name[64];
					c99c_node_t *mnode;

					sprintf(name, "%c=%d", m+'a', ((c99c_node_t *)ctx->map.map[m].straight)->uid);
					mnode = C99Nalloc(ctx, T_SYMLINK, c99c_strclone(name), NULL);
					C99Nchappend(ctx, lnk, mnode);
					mnode->parent = lnk;
				}
			}

		}
		gentrex_free(l->script);

		scr->parent = res_root;
		C99Nchappend(ctx, res_root, scr);
	}
}

